@@ -1,13 +1,43 @@
+# fmt: off
+import math
+
 import numpy as np
 import pytest
 import torch
+import torch.nn as nn
 
-from qsparse import (
-    prune,
-    structured_prune_callback,
-    unstructured_prune_callback,
-    unstructured_uniform_prune_callback,
-)
+from qsparse import prune, convert
+from qsparse.sparse import (MagnitudePruningCallback, UniformPruningCallback, 
+                            devise_layerwise_pruning_schedule, PruneLayer)
+
+# fmt: on
+
+def get_sparsity(tensor: torch.Tensor):
+    nz = tensor.nonzero()
+    if isinstance(nz, tuple):
+        nz = nz[0]
+    return 1 - len(nz) / np.prod(tensor.shape)
+
+
+class LeNet(nn.Module):
+    def __init__(self):
+        super(LeNet, self).__init__()
+        self.conv1 = nn.Conv2d(3, 6, kernel_size=5)
+        self.conv2 = nn.Conv2d(6, 16, kernel_size=5)
+        self.fc1 = nn.Linear(16 * 5 * 5, 120)
+        self.fc2 = nn.Linear(120, 84)
+        self.fc3 = nn.Linear(84, 10)
+
+    def forward(self, x):
+        x = F.relu(self.conv1(x))
+        x = F.max_pool2d(x, 2)
+        x = F.relu(self.conv2(x))
+        x = F.max_pool2d(x, 2)
+        x = x.view(x.size(0), -1)
+        x = F.relu(self.fc1(x))
+        x = F.relu(self.fc2(x))
+        x = self.fc3(x)
+        return x
 
 
 def test_feature():
@@ -24,6 +54,24 @@ def test_feature():
     assert np.isclose(get_sparsity(output), 0.5, atol=1 / np.prod(output.shape))
     assert np.all((output == 0) == (prune_layer.mask.numpy() == 0))
 
+    prune_layer = prune(
+        sparsity=0.5,
+        start=start,
+        interval=interval,
+        repetition=repetition,
+        dimensions={0,1,2,3},
+        callback=UniformPruningCallback(),
+    )
+    prune_layer.train()
+    for _ in range(
+        start + interval * (repetition + 2)
+    ):  # ensure the sparsification has been triggered
+        if _ == start + interval * (repetition + 2) - 1:
+            prune_layer.eval()
+        output = prune_layer(data).numpy()
+    assert np.isclose(get_sparsity(output), 0.5, atol=1 / np.prod(output.shape))
+    assert np.all((output == 0) == (prune_layer.mask.numpy() == 0))
+
     with pytest.raises(RuntimeError):
         prune_layer.eval()
         prune_layer(data2x)  # input shape shall be stationary
@@ -34,7 +82,7 @@ def test_feature():
         start=start,
         interval=interval,
         repetition=repetition,
-        strict=False,
+        dimensions={1}
     )
     for _ in range(
         start + interval * (repetition + 1)
@@ -48,13 +96,6 @@ def test_feature():
     )
 
 
-def get_sparsity(tensor: torch.Tensor):
-    nz = tensor.nonzero()
-    if isinstance(nz, tuple):
-        nz = nz[0]
-    return 1 - len(nz) / np.prod(tensor.shape)
-
-
 def test_weight():
     start, interval, repetition = 5, 2, 3
 
@@ -65,6 +106,7 @@ def test_weight():
         start=start,
         interval=interval,
         repetition=repetition,
+        callback=MagnitudePruningCallback(running_average=False),
     )
     pconv.train()
     for _ in range(
@@ -77,11 +119,32 @@ def test_weight():
     assert np.all(
         (pconv.weight.detach().numpy() == 0) == (pconv.prune.mask.detach().numpy() == 0)
     )
-
     assert not np.isclose(
         get_sparsity(dict(pconv.named_parameters())["weight"]), 0.5, atol=0.1
     ), "parameter['weight'] shall store the untouched weight without pruning"
 
+    pconv = prune(
+        torch.nn.Conv2d(10, 30, 3),
+        sparsity=0.5,
+        start=start,
+        interval=interval,
+        repetition=repetition,
+        callback=UniformPruningCallback(),
+    )
+    pconv.train()
+    for _ in range(
+        start + interval * (repetition + 1)
+    ):  # ensure the sparsification has been triggered
+        pconv(data)
+        print(pconv.prune._n_updates.item())
+    assert np.isclose(
+        get_sparsity(pconv.weight), 0.5, atol=1 / np.prod(pconv.weight.shape)
+    )
+    pconv.eval()
+    assert np.all(
+        (pconv.weight.detach().numpy() == 0) == (pconv.prune.mask.detach().numpy() == 0)
+    )
+
     pconv = prune(
         torch.nn.Conv2d(10, 30, 3),
         sparsity=0.5,
@@ -102,36 +165,38 @@ def test_weight():
         prune(torch.rand((10,)))
 
 
-def test_callback():
-    sparsity = 0.5
 
-    # unstructured
-    for shape in [(3, 32, 32), (3, 32), (32)]:
-        mask = unstructured_prune_callback(
-            [torch.rand(shape) for _ in range(10)], sparsity
-        )
-        result_sparsity = (~mask).sum() / np.prod(mask.shape)
-        assert np.isclose(sparsity, result_sparsity, atol=1 / np.prod(mask.shape))
+def test_more_pruning_options():
+    # gradient magnitude
+    shape = (3, 24, 24)
+    mean = torch.rand(*shape)
+    mask = torch.ones((1,) + shape)
+    layer = MagnitudePruningCallback(use_gradient=True)
+    for _ in range(1500):
+        inp = torch.normal(mean, 1)
+        inp = inp.view(1, *inp.shape)
+        inp.requires_grad = True
+        out = layer(inp, 0.5, mask)
+        out.backward(torch.rand((1,) + shape) / 10)
+    layer.eval()
+    assert np.isclose(get_sparsity(mask), 0.5, atol=2 / np.prod(shape))
 
-    for shape in [(3, 32, 32), (3, 32), (32)]:
-        mask = unstructured_uniform_prune_callback(
-            [torch.rand(shape) for _ in range(10)], sparsity
-        )
-        result_sparsity = (~mask).sum() / np.prod(mask.shape)
-        assert np.isclose(sparsity, result_sparsity, atol=1 / np.prod(mask.shape))
-    mask = unstructured_uniform_prune_callback(
-        [torch.rand(shape) for _ in range(10)], 0.7, current_mask=mask
+
+    mask = torch.ones(shape)
+    layer = MagnitudePruningCallback(l0=True)
+    for _ in range(1500):
+        inp = (torch.rand(*shape) > 0.5).float()
+        out = layer(inp, 0.5, mask)
+    assert np.isclose(get_sparsity(mask), 0.5, atol=2 / np.prod(shape))   
+
+
+def test_layerwise():
+    net = LeNet()
+    pnet = convert(
+        net,
+        prune(sparsity=0.5, callback=MagnitudePruningCallback()),
+        activation_layers=[nn.Conv2d, nn.Linear]
     )
-    result_sparsity = (~mask).sum() / np.prod(mask.shape)
-    assert np.isclose(0.7, result_sparsity, atol=1 / np.prod(mask.shape))
-
-    # structured
-    data = [torch.rand((32, 32, 32)) for _ in range(10)]
-    for channels in [{0}, {0, 1}, {0, 1, 2}]:
-        mask = structured_prune_callback(data, sparsity, prunable=channels)
-        for i in range(3):
-            if i not in channels:
-                assert mask.shape[i] == 1
-
-        result_sparsity = (~mask).sum() / np.prod(mask.shape)
-        assert np.isclose(sparsity, result_sparsity, atol=1 / np.prod(mask.shape))
+    pnet = devise_layerwise_pruning_schedule(pnet, start=10, interval=100, mask_refresh_interval=10)
+    sparsification_start_iters = [mod.start for mod in net.modules() if isinstance(mod, PruneLayer)]
+    assert sparsification_start_iters == sorted(sparsification_start_iters)
